"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DmlBuilderAdapter = void 0;
const _3xpr_1 = require("3xpr");
const typ3s_1 = require("typ3s");
const lambdaorm_base_1 = require("lambdaorm-base");
const domain_1 = require("../../../../query/domain");
const SqlString = require('sqlstring');
class DmlBuilderAdapter {
    // eslint-disable-next-line no-useless-constructor
    constructor(source, mapping, dialect, helper) {
        this.source = source;
        this.mapping = mapping;
        this.dialect = dialect;
        this.helper = helper;
    }
    build(sentence) {
        return new domain_1.Query({
            action: sentence.action,
            dialect: this.source.dialect,
            source: this.source.name,
            sentence: this.buildSentence(sentence),
            entity: sentence.entity,
            columns: sentence.columns,
            parameters: sentence.parameters,
            constraints: sentence.constraints,
            values: sentence.values,
            defaults: sentence.defaults
        });
    }
    buildSentence(sentence) {
        switch (sentence.crudAction) {
            case lambdaorm_base_1.SentenceCrudAction.select:
                return this.buildSelectSentence(sentence);
            case lambdaorm_base_1.SentenceCrudAction.insert:
                return this.buildInsertSentence(sentence);
            case lambdaorm_base_1.SentenceCrudAction.update:
                return this.buildUpdateSentence(sentence);
            case lambdaorm_base_1.SentenceCrudAction.delete:
                return this.buildDeleteSentence(sentence);
            default:
                throw new lambdaorm_base_1.SintaxisError(`sentence crud action ${sentence.crudAction} not found`);
        }
    }
    buildSelectSentence(sentence) {
        const map = sentence.children.find(p => p.name === 'map');
        const from = sentence.children.find(p => p instanceof lambdaorm_base_1.From);
        const joins = sentence.children.filter(p => p instanceof lambdaorm_base_1.Join);
        const filter = sentence.children.find(p => p.name === 'filter');
        const groupBy = sentence.children.find(p => p.name === 'groupBy');
        const having = sentence.children.find(p => p.name === 'having');
        const sort = sentence.children.find(p => p.name === 'sort');
        const page = sentence.children.find(p => p.name === 'page');
        const entity = this.mapping.getEntity(sentence.entity);
        if (entity === undefined) {
            throw new lambdaorm_base_1.SchemaError(`mapping undefined on ${sentence.entity} entity`);
        }
        if (map === undefined) {
            throw new lambdaorm_base_1.SchemaError('map operand not found');
        }
        if (from === undefined) {
            throw new lambdaorm_base_1.SchemaError('from operand not found');
        }
        let text = this.buildArrowFunction(map) + ' ' + this.buildFrom(from) + ' ' + this.buildJoins(entity, joins);
        if (filter) {
            text = text + this.buildArrowFunction(filter) + ' ';
        }
        if (groupBy) {
            text = text + this.buildArrowFunction(groupBy) + ' ';
        }
        if (having) {
            text = text + this.buildArrowFunction(having) + ' ';
        }
        if (sort) {
            text = text + this.buildArrowFunction(sort) + ' ';
        }
        if (page) {
            text = this.buildPage(text, page);
        }
        return text;
    }
    buildInsertSentence(sentence) {
        const insert = sentence.action === 'bulkInsert'
            ? sentence.children.find(p => p instanceof lambdaorm_base_1.BulkInsert)
            : sentence.children.find(p => p instanceof lambdaorm_base_1.Insert);
        const entity = this.mapping.getEntity(sentence.entity);
        if (entity === undefined) {
            throw new lambdaorm_base_1.SchemaError(`mapping undefined on ${sentence.entity} entity`);
        }
        if (insert === undefined) {
            throw new lambdaorm_base_1.SchemaError('insert operand not found');
        }
        return this.buildInsert(insert, entity);
    }
    buildUpdateSentence(sentence) {
        const update = sentence.children.find(p => p instanceof lambdaorm_base_1.Update);
        const filter = sentence.children.find(p => p.name === 'filter');
        const entity = this.mapping.getEntity(sentence.entity);
        if (entity === undefined) {
            throw new lambdaorm_base_1.SchemaError(`mapping undefined on ${sentence.entity} entity`);
        }
        if (update === undefined) {
            throw new lambdaorm_base_1.SchemaError('update operand not found');
        }
        let text = this.buildUpdate(update, entity);
        if (filter)
            text = text + this.buildArrowFunction(filter) + ' ';
        return text;
    }
    buildDeleteSentence(sentence) {
        const _delete = sentence.children.find(p => p instanceof lambdaorm_base_1.Delete);
        const filter = sentence.children.find(p => p.name === 'filter');
        const entity = this.mapping.getEntity(sentence.entity);
        if (entity === undefined) {
            throw new lambdaorm_base_1.SchemaError(`mapping undefined on ${sentence.entity} entity`);
        }
        if (_delete === undefined) {
            throw new lambdaorm_base_1.SchemaError('delete operand not found');
        }
        let text = this.buildDelete(_delete, entity);
        if (filter)
            text = text + this.buildArrowFunction(filter) + ' ';
        return text;
    }
    buildJoins(_entity, joins) {
        const list = [];
        const template = this.dialect.dml('join');
        for (const join of joins) {
            const entity = this.mapping.getEntity(join.name);
            if (entity === undefined) {
                throw new lambdaorm_base_1.SchemaError(`not found mapping for ${join.name}`);
            }
            let joinText = this.helper.str.replace(template, '{name}', this.dialect.delimiter(entity.mapping));
            joinText = this.helper.str.replace(joinText, '{alias}', join.alias);
            joinText = this.helper.str.replace(joinText, '{relation}', this.buildOperand(join.children[0])).trim();
            list.push(joinText);
        }
        return list.join(' ') + ' ';
    }
    buildFrom(from) {
        let template = this.dialect.dml('from');
        const entityMapping = this.mapping.entityMapping(from.entity);
        if (entityMapping === undefined) {
            throw new lambdaorm_base_1.SchemaError(`not found mapping for ${from.entity}`);
        }
        template = this.helper.str.replace(template, '{name}', this.dialect.delimiter(entityMapping));
        template = this.helper.str.replace(template, '{alias}', from.alias);
        return template.trim();
    }
    buildInsert(operand, entity) {
        let template = this.dialect.dml(operand instanceof lambdaorm_base_1.BulkInsert ? 'bulkInsert' : 'insert');
        const templateColumn = this.dialect.other('column');
        const fields = [];
        const values = [];
        const autoIncrement = this.mapping.getAutoIncrement(entity.name);
        if (autoIncrement && entity.sequence) {
            const templateSequenceNextVal = this.dialect.other('sequenceNextVal');
            fields.push(this.helper.str.replace(templateColumn, '{name}', this.dialect.delimiter(autoIncrement.mapping)));
            values.push(this.helper.str.replace(templateSequenceNextVal, '{name}', entity.sequence));
        }
        if (operand.children[0].type === _3xpr_1.OperandType.Obj) {
            const obj = operand.children[0];
            for (const p in obj.children) {
                const keyVal = obj.children[p];
                const property = entity.properties.find(q => q.name === keyVal.name);
                let name;
                if (property) {
                    name = property.mapping;
                }
                else {
                    name = keyVal.name;
                }
                fields.push(this.helper.str.replace(templateColumn, '{name}', this.dialect.delimiter(name)));
                values.push(this.buildOperand(keyVal.children[0]));
            }
        }
        template = this.helper.str.replace(template, '{name}', this.dialect.delimiter(entity.mapping));
        template = this.helper.str.replace(template, '{fields}', fields.join(','));
        template = this.helper.str.replace(template, '{values}', values.join(','));
        template = this.helper.str.replace(template, '{autoIncrementField}', autoIncrement && autoIncrement.mapping ? autoIncrement.mapping : '0');
        return template.trim();
    }
    buildUpdate(operand, entity) {
        let template = this.dialect.dml('update');
        const templateColumn = this.dialect.other('column');
        const templateAssign = this.dialect.operator('=', 2);
        const assigns = [];
        if (operand.children[0] instanceof Object) {
            const obj = operand.children[0];
            for (const p in obj.children) {
                const keyVal = obj.children[p];
                let name;
                const property = entity.properties.find(q => q.name === keyVal.name);
                if (property) {
                    name = property.mapping;
                }
                else {
                    name = keyVal.name;
                }
                const column = this.helper.str.replace(templateColumn, '{name}', this.dialect.delimiter(name));
                const value = this.buildOperand(keyVal.children[0]);
                let assign = this.helper.str.replace(templateAssign, '{0}', column);
                assign = this.helper.str.replace(assign, '{1}', value);
                assigns.push(assign);
            }
        }
        template = this.helper.str.replace(template, '{name}', this.dialect.delimiter(entity.mapping));
        template = this.helper.str.replace(template, '{alias}', operand.alias);
        template = this.helper.str.replace(template, '{assigns}', assigns.join(','));
        return template.trim() + ' ';
    }
    buildDelete(operand, entity) {
        let template = this.dialect.dml('delete');
        template = this.helper.str.replace(template, '{name}', this.dialect.delimiter(entity.mapping));
        template = this.helper.str.replace(template, '{alias}', operand.alias);
        return template.trim() + ' ';
    }
    buildPage(sentence, operand) {
        let template = this.dialect.dml('page');
        let page = 1;
        let records = 10;
        if (operand.children.length === 2) {
            page = parseInt(operand.children[0].name);
            records = parseInt(operand.children[1].name);
        }
        else if (operand.children.length === 3) {
            page = parseInt(operand.children[1].name);
            records = parseInt(operand.children[2].name);
        }
        if (page < 1)
            page = 1;
        template = this.helper.str.replace(template, '{sentence}', sentence);
        template = this.helper.str.replace(template, '{offset}', ((page - 1) * records).toString());
        template = this.helper.str.replace(template, '{records}', records.toString());
        return template.trim() + ' ';
    }
    buildOperand(operand) {
        if (operand instanceof lambdaorm_base_1.Sentence) {
            return this.buildSentence(operand);
        }
        else if (operand.type === _3xpr_1.OperandType.Arrow) {
            return this.buildArrowFunction(operand);
        }
        else if (operand.type === _3xpr_1.OperandType.CallFunc || operand.type === _3xpr_1.OperandType.ChildFunc) {
            return this.buildFunctionRef(operand);
        }
        else if (operand.type === _3xpr_1.OperandType.Operator) {
            return this.buildOperator(operand);
        }
        else if (operand.type === _3xpr_1.OperandType.Block) {
            return this.buildBlock(operand);
        }
        else if (operand.type === _3xpr_1.OperandType.Obj) {
            return this.buildObject(operand);
        }
        else if (operand.type === _3xpr_1.OperandType.List) {
            return this.buildList(operand);
        }
        else if (operand.type === _3xpr_1.OperandType.KeyVal) {
            return this.buildKeyValue(operand);
        }
        else if (operand instanceof lambdaorm_base_1.Field) {
            return this.buildField(operand);
        }
        else if (operand.type === _3xpr_1.OperandType.Var) {
            return this.buildVariable(operand);
        }
        else if (operand.type === _3xpr_1.OperandType.Const) {
            return this.buildConstant(operand);
        }
        else {
            throw new lambdaorm_base_1.SintaxisError(`Operand ${operand.type} ${operand.name} not supported`);
        }
    }
    buildArrowFunction(operand) {
        let template = this.dialect.dml(operand.name);
        for (let i = 0; i < operand.children.length; i++) {
            const text = this.buildOperand(operand.children[i]);
            template = this.helper.str.replace(template, '{' + i + '}', text);
        }
        return template.trim();
    }
    buildFunctionRef(operand) {
        const funcData = this.dialect.function(operand.name);
        if (!funcData)
            throw new lambdaorm_base_1.SintaxisError('Function ' + operand.name + ' not found');
        let text = '';
        if (['startWith', 'startsWith', 'like'].includes(operand.name) && operand.children.length === 2) {
            text = funcData.template;
            const firstOperand = this.buildOperand(operand.children[0]);
            const secondOperand = this.helper.str.replace(this.buildOperand(operand.children[1]), '\'', '');
            text = this.helper.str.replace(text, '{0}', firstOperand);
            text = this.helper.str.replace(text, '{1}', secondOperand);
        }
        else if (funcData.type === 'multiple') {
            const template = funcData.template;
            text = this.buildOperand(operand.children[0]);
            for (let i = 1; i < operand.children.length; i++) {
                text = this.helper.str.replace(template, '{accumulated}', text);
                text = this.helper.str.replace(text, '{value}', this.buildOperand(operand.children[i]));
            }
        }
        else {
            text = funcData.template;
            for (let i = 0; i < operand.children.length; i++) {
                const value = this.buildOperand(operand.children[i]);
                text = this.helper.str.replace(text, '{' + i + '}', value);
            }
        }
        return text;
    }
    buildOperator(operand) {
        let text = this.dialect.operator(operand.name, operand.children.length);
        for (let i = 0; i < operand.children.length; i++) {
            text = this.helper.str.replace(text, '{' + i + '}', this.buildOperand(operand.children[i]));
        }
        return text;
    }
    buildBlock(operand) {
        let text = '';
        for (const child of operand.children) {
            text += (this.buildOperand(child) + '');
        }
        return text;
    }
    buildObject(operand) {
        let text = '';
        const template = this.dialect.function('as').template;
        for (let i = 0; i < operand.children.length; i++) {
            const value = this.buildOperand(operand.children[i]);
            const forceDelimiter = ['PostgreSQL', 'Oracle'].includes(this.dialect.name);
            const alias = this.dialect.delimiter(operand.children[i].name, forceDelimiter);
            let fieldText = this.helper.str.replace(template, '{value}', value);
            fieldText = this.helper.str.replace(fieldText, '{alias}', alias);
            text += (i > 0 ? ', ' : '') + fieldText;
        }
        return text;
    }
    buildList(operand) {
        let text = '';
        for (let i = 0; i < operand.children.length; i++) {
            text += (i > 0 ? ', ' : '') + this.buildOperand(operand.children[i]);
        }
        return text;
    }
    buildKeyValue(operand) {
        return this.buildOperand(operand.children[0]);
    }
    buildField(field) {
        if (this.mapping.existsProperty(field.entity, field.fieldName())) {
            const property = this.mapping.getProperty(field.entity, field.fieldName());
            if (field.alias === undefined) {
                return this.helper.str.replace(this.dialect.other('column'), '{name}', this.dialect.delimiter(property.mapping, true));
            }
            else {
                let text = this.dialect.other('field');
                text = this.helper.str.replace(text, '{entityAlias}', field.alias);
                text = this.helper.str.replace(text, '{name}', this.dialect.delimiter(property.mapping));
                return text;
            }
        }
        else {
            const forceDelimiter = ['PostgreSQL', 'Oracle'].includes(this.dialect.name);
            return this.helper.str.replace(this.dialect.other('column'), '{name}', this.dialect.delimiter(field.name, forceDelimiter));
        }
    }
    buildVariable(operand) {
        const number = operand.number ? operand.number : 0;
        let text = this.dialect.other('variable');
        text = this.helper.str.replace(text, '{name}', this.helper.query.transformParameter(operand.name));
        text = this.helper.str.replace(text, '{number}', number.toString());
        return text;
    }
    buildConstant(operand) {
        if (operand.returnType === undefined) {
            return SqlString.escape(operand.name);
        }
        switch (operand.returnType.primitive) {
            case typ3s_1.Primitive.string:
                return SqlString.escape(operand.name);
            case typ3s_1.Primitive.boolean:
                return this.dialect.other(operand.name.toString());
            case typ3s_1.Primitive.integer:
                return parseInt(operand.name).toString();
            case typ3s_1.Primitive.number:
            case typ3s_1.Primitive.decimal:
                return parseFloat(operand.name).toString();
            default:
                return SqlString.escape(operand.name);
        }
    }
}
exports.DmlBuilderAdapter = DmlBuilderAdapter;
//# sourceMappingURL=dmlBuilder.js.map